//////////////////////////////////////////////////////////////////////////////
// Title: Class for Fourier Analysis
// Author: Chad Juliano
//////////////////////////////////////////////////////////////////////////////

package JavaScope;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import javax.swing.*;

/**
 * This class contains a set of static functions for Fourier analysis
 * applications and is not meant to be instantiated.
 * It is specialized for an application I am working on
 * but is still useful for other applications. It was designed to process
 * audio from the {@link SignalClip} class but can be used independently.
 * <p>
 * I am an Electrical Engineering student and I wrote this as part of a design project.
 * The latest version is available at
 * <a href="http://www.seas.smu.edu/~cjuliano/peruna.html">http://www.seas.smu.edu/~cjuliano/peruna.html</a>.
 * Feel free to use this without restriction as long as it is distributed unmodified.
 * If you find it useful or have questions, please send me e-mail at
 * <a href="chadj@pobox.com">chadj@pobox.com</a>. If I hear people are using this, it will
 * give me motivation to improve it.
 */



public class Fourier

{
    /** signal[REAL] represents the real part of a complex vector. */
    public static final int REAL=0;
    /** signal[IMAG] represents the imaginary part of a complex vector. */
    public static final int IMAG=1;
    public static void main(String[] args) {
  
        int N = 4;
        float[] A = new float[N];
        float[] B = new float[N];
        A[0] = 1;
        A[1] = 1;
        A[2] = 1;
        A[3] = 1;
        goertzelMag(A,B);
        System.out.print("gzl: ");
        printReal(B);
        DFTMag(A,B, LUT(A.length));
        System.out.print("DFT: ");
        printReal(B);
        FFTMag(A);
        System.out.print("FFT: ");
        printReal(A);
    }
    
    /**
     * Print a vector containing complex values
     */
    public static void printCplx(float[][] in) {
        DecimalFormat df = (DecimalFormat)NumberFormat.getInstance();
        df.setMaximumFractionDigits(2);
        for(int i=0; i<in[REAL].length; i++) {
            System.out.print("(" +  df.format(in[REAL][i])
            + ", " + df.format(in[IMAG][i]) + ") ");
        }
        System.out.print("\n");
    }
    /**
     * Print a vector containing real values.
     */
    public static void printReal(float[] in) {
        DecimalFormat df = (DecimalFormat)NumberFormat.getInstance();
        df.setMaximumFractionDigits(2);
        for(int i=0; i<in.length; i++) {
            System.out.print("(" + df.format(in[i]) + ") ");
        }
        System.out.print("\n");
    }
    /**
     * Compute the Discrete Fourier Transform of a complex vector.
     * @param in The complex vector containing the data to transform
     * @param out The complex vector to store the result in
     * @param lut a lookup table generated by {@link #LUT(int)}
     */
    public static void DFT(float[][] in, float[][] out, float[][] lut) {
        float[] rIn = in[REAL];
        float[] iIn = in[IMAG];
        float[] tCos = lut[REAL];
        float[] tSin = lut[IMAG];
        int N = in[REAL].length;
        float rAcc, iAcc;
        for(int k=0; k<N; k++) {
            rAcc = 0;
            iAcc = 0;
            for(int n=0, i=0; n<N; n++, i+=k) {
                if(i >= N) { i-=N; }
                rAcc += tCos[i]*rIn[n] - tSin[i]*iIn[n];
                iAcc += tCos[i]*iIn[n] + tSin[i]*rIn[n];
            }
            out[REAL][k] = rAcc;
            out[IMAG][k] = iAcc;
        }
    }
    /**
     * Compute the inverse of the DFT of a complex vector. Taking the DFT
     * followed by the IDFT results in the original vector.
     * @param in The complex vector containing the data to transform
     * @param out The complex vector to store the result in.
     * @param lut a lookup table generated by {@link #LUT(int)}
     */
    public static void IDFT(float[][] in, float[][] out, float[][] lut) {
        float[] rIn = in[REAL];
        float[] iIn = in[IMAG];
        float[] tCos = lut[REAL];
        float[] tSin = lut[IMAG];
        int N = in[REAL].length;
        float rAcc, iAcc;
        for(int k=0; k<N; k++) {
            rAcc = 0;
            iAcc = 0;
            for(int n=0, i=0; n<N; n++, i+=k) {
                if(i >= N) { i-=N; }
                rAcc += tCos[i]*rIn[n] + tSin[i]*iIn[n];
                iAcc += tCos[i]*iIn[n] - tSin[i]*rIn[n];
            }
            out[REAL][k] = rAcc / N;
            out[IMAG][k] = iAcc / N;
        }
    }
    /**
     * Compute the magnitude of the DFT of a real vector.
     * @param in The real vector containing the data to transform
     * @param out The real vector to store the result in.
     * @param lut a lookup table generated by {@link #LUT(int)}
     * @see #DFT(float[][], float[][], float[][])
     */
    public static void DFTMag(float[] in, float[] out, float[][] lut) {
        float[] tCos = lut[REAL];
        float[] tSin = lut[IMAG];
        int N = in.length;
        float rAcc, iAcc;
        for(int k=0; k<N; k++) {
            rAcc = 0;
            iAcc = 0;
            for(int n=0, i=0; n<N; n++, i+=k) {
                if(i >= N) { i-=N; }
                rAcc += tCos[i]*in[n];
                iAcc += tSin[i]*in[n];
            }
            out[k] = (float)Math.sqrt(rAcc*rAcc + iAcc*iAcc);
        }
    }
    
    /**
     * We needed this one for the spectragraph. It only computes half
     * the frequencies so no redundant information is computed.
     * @param in The real vector containing the data to transform
     * @param out The real vector to store the result in. This must be
     * half the length of the input.
     * @param lut a lookup table generated by {@link #LUT(int)}
     * @see #DFT(float[][], float[][], float[][])
     */
    public static void DFTMagHalf(float[] in, float[] out, float[][] lut) {
        float[] tCos = lut[REAL];
        float[] tSin = lut[IMAG];
        int N = in.length;
        int halfN = N/2;
        float rAcc, iAcc;
        float norm = 500/N;
        
        for(int k=0; k<halfN; k++) {
            rAcc = 0;
            iAcc = 0;
            for(int n=0, i=0; n<N; n++, i+=k) {
                if(i >= N) { i-=N; }
                rAcc += tCos[i]*in[n];
                iAcc += tSin[i]*in[n];
            }
            out[k] = (float)Math.sqrt(rAcc*rAcc + iAcc*iAcc)*norm;
        }
    }
    
    /**
     * Generate a lookup table for use with any DFT or IDFT related function.
     * A length N lookup table can only be used to transform length N vectors.
     * Using a length N table with data of a length other then N will give
     * you garbage.
     * @param N the length of the lookup table to generate
     * @return a sine and cosine table of length N
     */
    public static float[][] LUT(int N) {
        float[] tCos = new float[N];
        float[] tSin = new float[N];
        double Q = (-2*(float)Math.PI)/N;
        for(int i=0; i<N; i++) {
            tCos[i] = (float)Math.cos(Q*i);
            tSin[i] = (float)Math.sin(Q*i);
        }
        float[][] ret = { tCos, tSin };
        return ret;
    }
    
    /**
     * This is an implimentation of the second order Goertzel. Its
     * output is identical to the DFT but it requires no lookup
     * table and does half the multiplications; however, it is less
     * accurate.
     * @param in The complex vector containing the data to transform
     * @param out The complex vector to store the result in.
     */
    
    public static void goertzel(float[][] in, float[][] out) {
        float[] rIn = in[REAL];
        float[] iIn = in[IMAG];
        int N = in[REAL].length;
        float sin, cos, real1, real2, imag1, imag2, temp, twoCos;
        
        for(int i=0; i < N; i++) {
            sin = (float)Math.sin(2*Math.PI*i/N);
            cos = (float)Math.cos(2*Math.PI*i/N);
            real1 = rIn[1];
            real2 = 0;
            imag1 = iIn[1];
            imag2 = 0;
            twoCos = 2*cos;
            
            for(int j=1; j < N; j++) {
                temp = real1;
                real1 = twoCos*real1 - real2 + rIn[j];
                real2 = temp;
                temp = imag1;
                imag1 = twoCos*imag1 - imag2 + iIn[j];
                imag2 = temp;
            }
            out[REAL][i] = cos*real1 - sin*imag1 - real2;
            out[IMAG][i] = sin*real1 + cos*imag1 - imag2;
        }
    }
    
    /**
     * Calculate only the magnitude of the DFT of a real vector. This
     * algorithm is base on the second order Goertzel for
     * real vectors.
     * @param in The real vector containing the data to transform
     * @param out The real vector to store the result in. Values
     * are in the range [0, sqrt(2)] (assuming the input was
     * in the range [-1, 1])
     */
    public static void goertzelMag(float[] in, float[] out) {
        int N = in.length;
        float sin, cos, real1, real2, temp, twoCos;
        float fReal, fImag;
        
        for(int i=0; i < N; i++) {
            sin = (float)Math.sin(2*Math.PI*i/N);
            cos = (float)Math.cos(2*Math.PI*i/N);
            real1 = in[1];
            real2 = 0;
            twoCos = 2*cos;
            
            for(int j=1; j < N; j++) {
                temp = real1;
                real1 = twoCos*real1 - real2 + in[j];
                real2 = temp;
            }
            
            fReal = cos*real1 - real2;
            fImag = sin*real1;
            out[i] = (float)Math.sqrt(fReal*fReal + fImag*fImag);
        }
    }
    
    /**
     * Calculate only the phase angle of the DFT of a real vector.
     * This algorithm is based on the second order Goertzel for
     * real vectors.
     * @param in The real vector containing the data to transform
     * @param out The real vector to store the result in.
     * Values in the range [-PI, PI]
     * @param lut a lookup table generated by {@link #LUT(int)}
     */
    public static void goertzelPhase(float[] in, float[] out) {
        int N = in.length;
        float sin, cos, real1, real2, temp, twoCos;
        float fReal, fImag;
        
        for(int i=0; i < N; i++) {
            sin = (float)Math.sin(2*Math.PI*i/N);
            cos = (float)Math.cos(2*Math.PI*i/N);
            real1 = in[1];
            real2 = 0;
            twoCos = 2*cos;
            
            for(int j=1; j < N; j++) {
                temp = real1;
                real1 = twoCos*real1 - real2 + in[j];
                real2 = temp;
            }
            
            fReal = cos*real1 - real2;
            fImag = sin*real1;
            out[i] = (float)Math.atan(fImag/fReal);
        }
    }
    
    /**
     * In-place computation of the digit reversal.
     * The output of an FFT is in a special order known as digit reversal.
     * This will reorder the output so its order would be that of the DFT.
     * @see #radix2FFT(float[][])
     */
    public static void digitReverse(float[][] in) {
        int j=0;
        int k;
        int len = in[REAL].length;
        float[] real = in[REAL];
        float[] imag = in[IMAG];
        float temp;
        
        for(int i=1; i < len - 1; i++) {
            
            // generate the next bit reversed index
            k = len >> 1;
            while(k <= j) {
                j -= k;
                k >>= 1;
            }
            j += k;
            
            // j and i should now be a bit reversed permutation
            // Since each permutation has a mirror of itself,
            // we only swap one of the two.
            if(i < j) {
                temp = real[j];
                real[j] = real[i];
                real[i] = temp;
                
                temp = imag[j];
                imag[j] = imag[i];
                imag[i] = temp;
            }
        }
    }
    
    /**
     * In-place, radix-2, FFT with twiddle factor. As with
     * any FFT the output is digit reversed.
     * @param in must be a complex vector with the length being a power of 2
     * @see #digitReverse(float[][])
     */
    public static void radix2FFT(float[][] in) {
        float[] cRe = in[REAL];
        float[] cIm = in[IMAG];
        float E, C, S, Cl, Sl, xTemp, yTemp;
        int N = cRe.length;
        int N2 = N, N1;
        
        while(N2 > 1) {
            // main loop gets executed once per stage
            N1 = N2;
            N2 >>= 1;
            
            xTemp = (float)(2*Math.PI/(double)N1);
            C = 1;
            S = 0;
            
            // yes, these are slow but because we use twiddle
            // factors they are only called once per stage
            // which is insignificant
            Cl = (float)Math.cos(xTemp);
            Sl = (float)Math.sin(xTemp);
            
            for(int j=0; j<N2; j++) {
                for(int i=j, l=j+N2; i<N; i+=N1, l+=N1) {
                    // do a butterfly at i and l
                    xTemp = cRe[i] - cRe[l];
                    cRe[i] += cRe[l];
                    yTemp = cIm[i] - cIm[l];
                    cIm[i] += cIm[l];
                    cRe[l] = C*xTemp + S*yTemp;
                    cIm[l] = C*yTemp - S*xTemp;
                }
                
                // update the twiddle factors
                xTemp = C;
                C = C*Cl - S*Sl;
                S = xTemp*Sl + S*Cl;
            }
        }
    }
    
    /**
     * Compute the magnitude of the DFT.
     * The function is similar to {@link #goertzelMag(float[],float[])} except
     * for the power-of-2 limitation and in-place computation.
     * @param in must be have a power of 2 length
     * @see #radix2FFT(float[][] in)
     */
    public static void FFTMag(float[] in) {
        float[] imag = new float[in.length];
        float [][] cplx = { in, imag };
        radix2FFT(cplx);
        digitReverse(cplx);
        
        float rTemp, iTemp;
        for(int i=0; i < in.length; i++) {
            rTemp = in[i];
            iTemp = imag[i];
            in[i] = (float)Math.sqrt(rTemp*rTemp + iTemp*iTemp);
        }
    }
    
    /**
     * Compute the phase of the DFT.
     * The function is similar to {@link #goertzelPhase(float[],float[])} except
     * for the power-of-2 limitation and in-place computation.
     * @param in must be have a power of 2 length
     * @see #radix2FFT(float[][] in)
     */
    public static void FFTPhase(float[] in) {
        float[] imag = new float[in.length];
        float [][] cplx = { in, imag };
        radix2FFT(cplx);
        digitReverse(cplx);
        
        float rTemp, iTemp;
        for(int i=0; i < in.length; i++) {
            rTemp = in[i];
            iTemp = imag[i];
            in[i] = (float)Math.atan(iTemp/rTemp);
        }
    }
    
    /**
     * This is used to compute a DFT of one length based upon a real
     * vector of another length. A typical use case is as follows:
     * <p>
     * <ol>
     * <li> Create the output vector with the same length as the desired DFT.
     * <li> Call resample(in, out) with the input and output vectors.
     * <li> Use the output vector as real data in the calculation of the DFT.
     * </ol>
     * <p>
     * If the output is greater then the input, the result is zero padded.
     * If the input is greater then the output, the result is circular aliased.
     * @param in The time domain vector to down sample in the Fourier domain
     * @param out The result in the time domain
     * @see #DFT(float[][], float[][], float[][])
     * @see #resample(float[], int, int, float[], int, int)
     */
    public static void resample(float[] in, float[] out) {
        resample(in, 0, in.length, out, 0, out.length);
    }
    
    /**
     * Resample a segment of a vector into a segment of an output
     * vector.
     * @param in the input array
     * @param inStart the starting index into the input
     * @param inLen the number of samples after inStart
     * @param out the output array
     * @param outStart the starting index into the output
     * @param outLen the number of samples after outStart
     * @see #resample(float[], float[])
     */
    public static void resample(float[] in, int inStart, int inLen,
    float[] out, int outStart, int outLen) {
        System.arraycopy(in, inStart, out, outStart, Math.min(inLen, outLen));
        
        if(inLen <= outLen) {
            // pad out with zeros
            int end = outStart + outLen;
            for(int i=outStart + inLen; i<end; i++) {
                out[i] = 0;
            }
            return;
        }
        
        // do the circular alias and normalize
        float acc;
        int end;
        float nFact = (float)inLen/(float)outLen;
        for(int i=0; i<outLen; i++) {
            acc = out[i + outStart];
            end = inStart+inLen;
            for(int j=i+inStart+outLen; j<end; j+=outLen) {
                acc += in[j];
            }
            out[i + outStart] = acc*nFact;
        }
    }
    
    /**
     * Normalize the array so the largest value is 1.
     */
    public static void normalize(float[] in) {
        float max = Float.NEGATIVE_INFINITY;
        float val;
        
        for(int i=0; i < in.length; i++) {
            val = in[i];
            if(val < 0) { val = -val; }
            if(val > max) { max = val; }
        }
        
        for(int i=0; i < in.length; i++) {
            in[i] = in[i]/max;
        }
    }
    
    /**
     * Do direct convolution of H and X. If Hl = H.length and
     * Xl = X.length then it has runtime of approximately
     * O(Hl^2 + Hl*Xl). This means that things are more efficient
     * when H is shorter then X.
     * @param pBar a progress bar to update or null
     * @return a vector of length (H.length + X.length -1)
     */
    public static float[] convolve(float[] H, float[] X, JProgressBar pBar) {
        int length = H.length + X.length - 1;
        float[] out = new float[length];
        
        if(pBar != null) {
            pBar.setMinimum(0);
            pBar.setMaximum(length - 1);
        }
        
        for(int i=0; i<out.length; i++) {
            if(i%256 == 0) {
                if(Thread.currentThread().isInterrupted()) { break; }
                if(pBar != null) { pBar.setValue(i); }
            }
            
            int index;
            float acc = 0;
            for(int j=0; j<H.length; j++) {
                index = i-j;
                if((index >= 0) && (index < X.length)) {
                    acc = acc + H[j]*X[index];
                }
            }
            out[i] = acc;
        }
        return out;
    }
    
    /**
     * Generate a sequence to implement a low-pass filter.
     * Convolution with this sequence will pass only
     * frequencies below the cutoff frequency.
     * @param fc the cutoff frequency in hz
     * @param fs the sampling frequency in hz
     * @see #highPass(int, int)
     * @see #convolve(float[], float[], JProgressBar)
     */
    public static float[] lowPass(int fc, int fs) {
        double rads = 2*Math.PI*fc/fs;
        float[] out = new float[500];
        int index;
        
        for(int i=0; i<out.length; i++) {
            index = i - out.length/2;
            out[i] = (float)(Math.sin(rads*index)/(Math.PI*index));
        }
        out[out.length/2] = (float)(rads/Math.PI);
        return out;
    }
    
    /**
     * Generate a sequence to implement a high-pass filter.
     * Convolution with this sequence will pass only
     * frequencies above the cutoff frequency.
     * @param fc the cutoff frequency in hz
     * @param fs the sampling frequency in hz
     * @see #lowPass(int, int)
     * @see #convolve(float[], float[], JProgressBar)
     */
    public static float[] highPass(int fc, int fs) {
        float[] out = lowPass(fs/2 - fc, fs);
        for(int i = 1; i < out.length; i+=2) {
            out[i] = -out[i];
        }
        return out;
    }
    public static float[] hanning(int nc){
        
        
        float[]   window = new float[nc];
        for(int i = 0; i < nc; i++) {
            // hanning window
            window[i] = (float)(0.5*(1.0 - Math.cos(2*Math.PI*i/(nc - 1))));
        }
        return window;
    }
    public static float[] hamming(int nc){
        float[]    window = new float[nc];
        for(int i = 0; i < nc; i++) {
            // create the hamming window
            window[i] = (float)(0.54 - 0.46*   Math.cos(2*Math.PI*i/(nc - 1)));
            
        }
        return window;
    }
    public static float[] rect(int nc){
        float[]    window = new float[nc];
        for(int i = 0; i < nc; i++) {
            window[i] = (float)(1);
        }
        return window;
    }
    
    
}